#!/usr/bin/env python

import curses, string, traceback
from curses import ascii
from curses.textpad import rectangle

DEBUG = 0
TK = 0

# Constants
SOFT = 0
HARD = 1

# TODO: 
# backfill if stuff is deleted from a line

if TK:
	from Tkinter import *
	import Pmw
	class Logger(Frame):
		def __init__(self):
			Frame.__init__(self)
			self.pack()
			self.text = Pmw.ScrolledText()
			self.text.pack(side=TOP, expand=1, fill=BOTH)
			self.update_idletasks()
		def log(self, msg):
			self.text.insert('end', msg + '\n')
			self.text.see('end')
			self.update_idletasks()
	if DEBUG:
		global LoggerInstance
		LoggerInstance = Logger()

def LOG(msg):
	if DEBUG:
		textLOG(msg)
		if TK: LoggerInstance.log(msg)

def textLOG(msg):
	if DEBUG:
		f = open('log', 'a')
		f.write(msg + '\n')
		f.close()

ErrorHandler = None
def setErrorHandler(f):
	global ErrorHandler
	ErrorHandler = f

def ERROR(msg):
	if ErrorHandler is not None:
		ErrorHandler('tpad error: ' + msg)
	f = open('log', 'a')
	traceback.print_exc(None, f)
	f.close()

class Tpad:
	def __init__(self, root, standalone=0):
		self.standalone = standalone
		curses.raw()
		self.win = root
		self.maxy, self.maxx = self.win.getmaxyx()
		LOG('maxx: %i, maxy: %i' % (self.maxx, self.maxy))
		self.win.keypad(1)
		self.lines = ['']
		self.breaks = [HARD]
		self.curline = 0
		self.pos = 0	# x coord of cursor on self.curline
		self.topVisible = 0		# index of first visible line

	def focus(self):
		self.win.move(self.curline - self.topVisible, self.pos)

	def getline(self, index=None):
		if index is None: index = self.curline
		return self.lines[index]
	
	def setline(self, buffer, index=None):
		if index is None: index = self.curline
		self.lines[index] = buffer

	def ensure_line(self, index=None):
		if index is None: index = self.curline
		while len(self.lines) < index + 1:
			self.lines.append('')
			self.breaks.append(HARD)

	def joinWithPrev(self, line=None):
		# joins "line" with the previous line
		# line is in buffer coordinates
		# does not update cursor, curline, or pos
		# if these change, they must be updated on return
		# this may trigger a word wrap --- thus this function returns the
		# (y, x) of the new cursor position after join
		if line is None:
			line = self.curline
		y = line - self.topVisible
		x = self.pos

		LOG('joining two lines')
		prevline = self.getline(line - 1)
		prevlen = len(prevline)
		thisline = self.getline(line)

		LOG('previous line is %s' % `prevline`)
		LOG('current line is %s' % `thisline`)
		prevline = prevline[:prevlen] + thisline
		self.setline(prevline, line - 1)
		del self.lines[line]
		del self.breaks[line-1]

		# move cursor to previous line at the junction point
		LOG('scrolling to previous line (y %i, x %i)'
			% (y-1, prevlen))
		newy, newx = self.scrollto(y-1, prevlen)
		# add the first part of this line on to the end of the previous
		# line (but only up to maxx total characters)
		LOG('addstr (%i, %i) %s' % (newy, newx,
			`thisline[:self.maxx-prevlen]`))
		self.win.addstr(newy, newx,
			thisline[:self.maxx-prevlen])
		line = line - 1

		# scroll the next lines up
		# from line y to self.maxy-1, inclusive
		LOG('testing scroll from %i to %i' % (newy+1, self.maxy-1))
		if newy+1 < self.maxy - 1:
			LOG('scrolling by 1 between %i and %i' %
				(newy+1, self.maxy-1))
			self.win.scrollok(1)
			LOG('setting scrreg')
			self.win.setscrreg(newy+1, self.maxy-1)
			self.win.scroll(1)
			LOG('setting scrreg again')
			self.win.setscrreg(0, self.maxy-1)
			self.win.scrollok(0)
		# now repaint the last line; erase it first
		# always need to repaint the last line because the former last
		# line will have scrolled up to become the next-to-last line
		lastlineindex = self.topVisible + self.maxy - 1
		LOG('moving to (%i, %i) and clearing to eol' % (self.maxy-1, 0))
		self.win.move(self.maxy-1, 0)
		self.win.clrtoeol()
		if lastlineindex < len(self.lines):
			LOG('repainting last line at (%i, 0)' %
				(self.maxy-1))
			self.win.addstr(self.maxy-1, 0, self.getline(
				self.topVisible + self.maxy - 1))

		y = line - self.topVisible
		x = prevlen
		y, x = self.scrollto(y, x)
		# wrap if necessary
		LOG('checking for wrap')
		if len(prevline) >= self.maxx:
			LOG('...wrapping')
			y, x = self.wrap(line, prevlen)
		else:
			LOG('...not wrapping')
		return (y, x)

	def wrap(self, startline, x):
		# startline is in buffer coords
		# y and x are the position of the cursor in screen coords
		# returns (y, x) of new cursor position
		y = startline - self.topVisible
		LOG('wrapping line %i (y %i, x %i)' % (startline, y, x))
		line = self.getline(startline)
		origlinelen = len(line)
		LOG('  contents %s, len %i' % (`line`, len(line)))
		breakafter = min(len(line)-1, self.maxx-2)
		while breakafter and not line[breakafter].isspace():
			breakafter = breakafter - 1
		LOG('last space at position %i' % breakafter)
		if breakafter == 0:
			# there is no space; do a soft break right at the end
			LOG('no space; soft wrapping at end')
			breakafter = self.maxx - 2
		newline = line[:breakafter+1]
		overflow = line[breakafter+1:]

		breaktype = self.breaks[startline]
		LOG('this line has a %s break' % (['SOFT', 'HARD'][breaktype]))
		self.setline(newline, startline)

		LOG('line %i will be broken after char #%i' % (y, breakafter))
		LOG('overflow is %s' % `overflow`)
		LOG('orig line len %i = new line len %i + overflow %i' % (
			origlinelen, len(newline), len(overflow)))
		if self.pos > breakafter:
			# the cursor position wrapped to the next line
			newy = y+1
			LOG('x is %i, len(newline) is %i' % (x, len(newline)))
			newx = x - len(newline)
			LOG('new position on line %i is %i' % (newy, newx))
		else:
			newy = y
			newx = x

		# get rid of it in the display
		LOG('Clearing to the right of that')
		if y < self.maxy:
			self.win.move(y, breakafter+1)
			self.win.clrtoeol()

		if breaktype == SOFT:
			# prepend onto next line
			self.ensure_line(startline + 1)

			LOG('inserting line %i' % (startline + 1))
			line = self.getline(startline + 1)
			line = overflow + line
			self.setline(line, startline + 1)
			LOG('next line is now %s' % `line`)
			if y+1 < self.maxy:
				self.win.insnstr(y+1, 0, overflow, 0)
		else:
			# create new line
			self.breaks[startline] = SOFT
			self.lines.insert(startline + 1, overflow)
			self.breaks.insert(startline + 1, HARD)
			# repaint
			if y+1 < self.maxy:
				self.win.move(y+1, 0)
				self.win.clrtobot()
				for i in range(y+1, min(self.maxy,
					len(self.lines)-self.topVisible)):
					LOG('repainting line %i' % i)
					LOG('contents: %s' %
						`self.getline(self.topVisible+i)`)
					self.win.addstr(i, 0,
						self.getline(self.topVisible + i)[:self.maxx])

		# wrap this line if necessary
		if len(self.getline(startline + 1)) >= self.maxx:
			self.wrap(startline+1, 0)
		LOG('after wrap, returning (y, x) of (%i, %i)' % (newy, newx))
		return (newy, newx)

	def scrollto(self, y, x):
		# x and y are in screen coordinates not buffer coordinates
		# returns the new y, x of the cursor in screen coords 
		# does not modify self.curline or self.pos
		LOG('scrolling to (y %i, x %i)' % (y, x))
		LOG('top visible: %i, curline: %i' % (self.topVisible,
			self.curline))
		if y < 0:
			LOG('scrolling up')
			self.win.scrollok(1)
			self.win.scroll(y)
			self.win.scrollok(0)
			self.topVisible = self.topVisible + y
			for i in range(-y):
				LOG('backfilling line %i: %s' % (self.topVisible + i,
					`self.getline(self.topVisible + i)`))
				LOG('at physical row %i' % i)
				s = self.getline(self.topVisible + i)
				LOG('addstr(%i, 0, %s)' % (i, `s`))
				self.win.addstr(i, 0, s)
			y = 0
		elif y >= self.maxy:
			# scroll down
			LOG('scrolling down')
			numlines = y - self.maxy + 1
			LOG('scrolling by %i' % numlines)
			bottomVisible = self.topVisible + self.maxy - 1
			self.win.scrollok(1)
			self.win.scroll(numlines)
			self.win.scrollok(0)
			# fill in lines
			# base is in buffer coordinates
			base = self.topVisible + self.maxy
			# index is in screen coordinates
			index = self.maxy - numlines
			LOG('base: %i, index %i' % (base, index))
			for i in range(y - self.maxy + 1):
				# TODO set coords
				LOG('looping over lines to add, i=%i' % i)
				LOG('forward fill printing line %i' % (base + i))
				self.ensure_line(base + i)
				LOG('  contents: %s' % `self.getline(base+i)`)
				LOG('  at line %i' % (index + i))
				self.win.addstr(index + i, 0,
					self.getline(base + i))
			self.topVisible = self.topVisible + (y-self.maxy+1)
			y = self.maxy - 1
		self.win.move(y, x)
		return y, x

	def mainloop(self):
		while 1:
			ch = self.win.getch()
			ret = self.safeprocess(ch)
			if ret: return

	def safeprocess(self, ch):
		try:
			ret = self.process(ch)
		except Exception, e:
			ERROR(str(e))
			ret = 0
		return ret
	
	def process(self, ch):
		# Return 1 iff processing this key terminates the editor
		key = curses.unctrl(ch)
		keyname = curses.keyname(ch)
		(y, x) = self.win.getyx()
		LOG('processing new command: %s %s' % (key, keyname))
		LOG('y %i, x %i, curline %i, topVis %i, pos %i' % (
			y, x, self.curline, self.topVisible, self.pos))
		if DEBUG:
			assert x == self.pos
			assert y == self.curline - self.topVisible
		if self.standalone and keyname == '^C':
			return 1
		# Dan says "if its seven-bit-masked value is printable, then it's
		# probably printable"
		if ascii.isprint(ch & 0x7f):
			# insert printable character
			LOG('printable %s, key %s, pos %i len line %i' % (`ch`,
				`key`, self.pos, len(self.getline())))
			if self.pos == len(self.getline()):
				# cursor is at the end of the line
				LOG('inserting character at end of line')
				# insert into buffer
				buffer = self.getline()
				buffer = buffer[:self.pos] + key + buffer[self.pos:]
				self.setline(buffer)

				self.pos = self.pos + 1

				if len(buffer) >= self.maxx:
					LOG('wrapping because it is too long')
					y, x = self.wrap(self.curline, x+1)
					self.curline = self.curline + 1
					self.pos = x
					LOG('finished wrap, curline %i, pos %i' % (
						self.curline, self.pos))
					self.scrollto(y, x)
				else:
					# update display
					self.win.addch(ch)

			else:
				# insert other than at end of line
				LOG('inserting in middle of line')
				self.win.insch(key)
				buffer = self.getline()
				buffer = buffer[:self.pos] + key +\
					buffer[self.pos:]
				self.setline(buffer)
				LOG('  contents of line: %s (len %i)' % (
					`buffer`, len(buffer)))
				if len(buffer) >= self.maxx:
					y, x = self.wrap(self.curline, x+1)
					# x and y are in screen coords
					self.pos = x
					self.curline = y + self.topVisible
				else:
					self.pos = self.pos + 1
				self.scrollto(self.curline - self.topVisible, self.pos)

		# ^b
		elif ch in [curses.KEY_LEFT, ascii.STX]:
			if self.pos == 0:
				# scroll to previous line
				if self.curline > 0:
					self.curline = self.curline - 1
					self.pos = len(self.getline())
			else:
				self.pos = self.pos - 1
			self.scrollto(self.curline - self.topVisible, self.pos)
		# ^f
		elif ch in [curses.KEY_RIGHT, ascii.ACK]:
			if x < len(self.getline()):
				self.pos = self.pos + 1
			else:
				if self.curline < len(self.lines) - 1:
					self.pos = 0
					self.curline = self.curline + 1
			self.scrollto(self.curline - self.topVisible, self.pos)
		# backspace
		elif ch in [ascii.BS, ascii.DEL, curses.KEY_BACKSPACE]:
			if x != 0 or self.curline != 0:
				brk = SOFT
				if x == 0 and self.curline > 0:
					# join with previous line
					LOG('pressed delete at beginning of line; joining')
					brk = self.breaks[self.curline-1]
					lenprev = len(self.getline(self.curline - 1))
					y, x = self.joinWithPrev(self.curline)
					self.curline = y + self.topVisible
					self.pos = x
				if brk == SOFT and self.pos > 0:
					buffer = self.getline()
					LOG('about to delete char from line')
					LOG('... Line is %s' % `buffer`)
					LOG('... pos is %i, curline is %i' % (self.pos,
						self.curline))
					buffer = buffer[:self.pos-1] +\
						buffer[self.pos:]
					self.setline(buffer)
					LOG('deleted character; line %i is now %s' % (
						self.curline, `buffer`))
					# move cursor left by one
					self.pos = self.pos - 1
					self.scrollto(self.curline - self.topVisible,
						self.pos)
					# delete the character under the cursor
					LOG('Deleting char at (%i, %i)' %
						(self.curline - self.topVisible,
						self.pos))
					self.win.delch()
				self.scrollto(self.curline - self.topVisible, self.pos)
		# ^d
		elif ch == ascii.EOT:
			buffer = self.getline()
			if self.pos == len(buffer) and \
				self.curline == len(self.lines) - 1:
				# end of buffer, return true
				return 1
			# delete a character
			LOG('deleting a character')
			if self.pos == len(buffer):
				# join with next; equivalent to ^k at this point
				y, x = self.joinWithPrev(self.curline+1)
				self.curline = y + self.topVisible
				self.pos = x
				# don't change current cursor position
			else:
				# deleting char from middle of line
				self.win.delch()
				buffer = buffer[:self.pos] + buffer[self.pos+1:]
				self.setline(buffer)
			self.scrollto(self.curline - self.topVisible, self.pos)
		# ^e
		elif ch in [ascii.ENQ, curses.KEY_LL]:
			# end of line
			self.pos = len(self.getline())
			self.scrollto(self.curline - self.topVisible, self.pos)
		# ^a
		elif ch in [ascii.SOH, curses.KEY_HOME]:
			# beginning of line
			self.pos = 0
			self.scrollto(self.curline - self.topVisible, self.pos)
		# ^n, down
		elif ch in [curses.KEY_DOWN, ascii.SO]:
			LOG('Received down arrow, curline is %i, # lines %i' %
				(self.curline, len(self.lines)))
			if self.curline < len(self.lines) - 1:
				self.curline = self.curline + 1
				self.ensure_line()
				lastx = len(self.getline())
				if lastx < self.pos:
					self.pos = lastx
				self.scrollto(self.curline - self.topVisible, self.pos)
		# ^p, up
		elif ch in [curses.KEY_UP, ascii.DLE]:
			if self.curline > 0:
				self.curline = self.curline - 1
				lastx = len(self.getline())
				if lastx < self.pos:
					self.pos = lastx
				self.scrollto(self.curline - self.topVisible, self.pos)
		# \n, newline
		elif ch == ascii.NL:
			# insert new line after current line
			line = self.getline()
			prev = line[:self.pos]
			next = line[self.pos:]
			curbreak = self.breaks[self.curline]
			self.setline(prev)
			self.win.clrtoeol()
			self.lines.insert(self.curline + 1, next)
			self.breaks.insert(self.curline + 1, curbreak)
			self.breaks[self.curline] = HARD

			if y+1 < self.maxy - 1:
				# if the scrolling region is more than one line
				# high
				self.win.scrollok(1)
				self.win.setscrreg(y+1, self.maxy-1)
				self.win.scroll(-1)
				self.win.setscrreg(0, self.maxy-1)
				self.win.scrollok(0)
			if y+1 <= self.maxy - 1:
				# if the next line is visible
				self.win.move(y+1, 0)
				self.win.clrtoeol()
				self.win.addstr(y+1, 0, next)

			self.curline = self.curline + 1
			y = y + 1
			self.pos = 0
			self.scrollto(self.curline - self.topVisible, self.pos)
		# ^u
		elif ch == ascii.NAK:
			self.setline('')
			self.pos = 0
			self.scrollto(self.curline - self.topVisible, self.pos)
			self.win.clrtoeol()
		# ^w
		elif ch == ascii.ETB:
			if x != 0 or self.curline != 0:
				if x == 0 and self.curline > 0:
					prevlen = len(self.getline(self.curline - 1))
					y, x = self.joinWithPrev(self.curline)
					self.curline = y + self.topVisible
					self.pos = x
				x = self.pos
				y = self.curline - self.topVisible
				savex = x
				# lop off trailing spaces
				while x > 0 and self.win.inch(y, x-1) == ascii.SP:
					self.scrollto(y, x-1)
					self.win.delch()
					x = x - 1
				# lop off the previous word
				while x > 0 and self.win.inch(y, x-1) != ascii.SP:
					self.scrollto(y, x-1)
					self.win.delch()
					x = x - 1
				self.scrollto(y, x)
				# update the buffer
				buffer = self.getline()
				buffer = buffer[:x] + buffer[savex:]
				self.setline(buffer)
				self.pos = x
				self.scrollto(self.curline - self.topVisible, self.pos)
		# page up
		elif ch == curses.KEY_PPAGE:
			if (self.curline - self.topVisible) == 0 and self.pos == 0:
				# scroll up a page
				numlines = min(self.topVisible, self.maxy - 1)
				LOG('scrolling by %i' % -numlines)
				self.scrollto(-numlines, 0)
				self.curline = self.curline - numlines
				LOG('curline now %i' % self.curline)
				self.pos = 0
			else:
				self.pos = 0
				self.curline = self.topVisible
				self.scrollto(0, 0)
		# page down
		elif ch == curses.KEY_NPAGE:
			if (self.curline - self.topVisible) == self.maxy - 1:
				# scroll down one page
				botVis = self.topVisible + self.maxy - 1
				numlines = min(len(self.lines)-botVis-1,
					self.maxy - 1)
				LOG('botVis is %i' % botVis)
				LOG('scrolling down %i lines' % numlines)
				self.curline = self.curline + numlines
				self.pos = len(self.getline())
				LOG('  content of line is %s' % `self.getline()`)
			else:
				maxline = min(len(self.lines) - self.topVisible - 1,
					self.maxy - 1)
				self.pos = len(self.getline(maxline + self.topVisible))
				LOG('scrolling down to y %i x %i' % (maxline, self.pos))
				LOG('last line has contents %s' % `self.getline(maxline
					+ self.topVisible)`)
				self.curline = self.curline + maxline
			self.scrollto(self.curline - self.topVisible, self.pos)
		# ^k
		elif ch == ascii.VT:
			buffer = self.getline()
			LOG('killing a line')
			if self.pos == len(buffer) and \
				self.curline < len(self.lines) - 1:
				LOG('... joining with next')
				y, x = self.joinWithPrev(self.curline+1)
				self.curline = y + self.topVisible
				self.pos = x
				# don't change current cursor position
			else:
				LOG('... but only to right of cursor')
				buffer = buffer[:self.pos]
				self.setline(buffer)
				self.win.clrtoeol()
			self.scrollto(self.curline - self.topVisible, self.pos)
		# default
		else:
			LOG('unrecognized command %s' % `key`)

	def get(self):
		LOG('%i lines' % len(self.lines))
		s = ''
		for i in range(len(self.lines)):
			s = s + self.lines[i]
			if self.breaks[i] == HARD:
				s = s + '\n'
		return unicode(s, 'latin-1')

def tpad(root):
	global DEBUG
	DEBUG = 1

	rectangle(root, 0, 0, 6, 41)
	root.refresh()
	win = root.subwin(5, 40, 1, 1)
	t = Tpad(win, standalone=1)
	t.mainloop()
	LOG(`t.get()`)
	return t.get()

def main():
	ret = curses.wrapper(tpad)
	print 'Value:'
	print `ret`
	print 'Or literally:'
	print ret.encode('latin-1')

if __name__ == '__main__':
	main()

